/*
  proto_map.cc

  $Id: proto_map.c,v 1.5 1999/12/06 01:53:07 everton Exp $
 */

#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "proto_map.hpp"
#include "forward.h"
#include "portfwd.h"

void proto_map::show() const
{
  iterator<vector<int>,int> it(*port_list);
  it.start();
  if (it.cont()) {
    syslog(LOG_INFO, "%d", it.get());
    it.next();
  }
  for (; it.cont(); it.next())
    syslog(LOG_INFO, ", %d", it.get());

  syslog(LOG_INFO, " /* uid: %d, gid: %d */", uid, gid);

  if (ftp_pasv)
    syslog(LOG_INFO, " ftp-passive-mode {\n");
  else
    syslog(LOG_INFO, " {\n");
  iterator<vector<host_map*>,host_map*> it2(*map_list);
  it2.start();
  if (it2.cont()) {
    it2.get()->show();
    it2.next();
  }
  for (; it2.cont(); it2.next()) {
    syslog(LOG_INFO, ";\n");
    it2.get()->show();
  }
  syslog(LOG_INFO, "\n}\n");
}

void grandchild_reaper(int sig)
{
  int status;

  pid_t gchild_pid = wait(&status);

  if (gchild_pid == -1) {
    syslog(LOG_ERR, "child: Error waiting pid of grandchild: %m");
    return;
  }
  
  for (int fd = 0; fd < MAX_FD; ++fd)
    if (grandchild_pid[fd] == gchild_pid)
      grandchild_pid[fd] = -1;

  if (WIFEXITED(status)) {
    ONVERBOSE(syslog(LOG_WARNING, "child: Grandchild with PID %d exited normally with exit status: %d\n", gchild_pid, WEXITSTATUS(status)));
    return;
  }

  if (WIFSIGNALED(status))
    syslog(LOG_NOTICE, "child: Grandchild received signal: %d\n", WTERMSIG(status)); 
    
  syslog(LOG_ERR, "child: Grandchild with PID %d exited abnormally\n", gchild_pid);
}

void proto_map::spawn_server(proto_t proto, int port) const
{
  pid_t pid = fork();
  if (pid) {
    if (pid < 0)
      syslog(LOG_ERR, "fork() failed: %m");

    return;
  }

  /* Child */

  signal(SIGCHLD, grandchild_reaper);

  switch (proto) {
  case P_TCP:
    tcp_forward(port, map_list, ftp_pasv, uid, gid);
    break;

  case P_UDP:
    udp_forward(port, map_list, uid, gid);
    break;

  default:
    syslog(LOG_ERR, "Unknown protocol identifier: %d\n", proto);
  }
    
  syslog(LOG_ERR, "Child exiting (!)\n");
  exit(1);
}

void proto_map::serve(proto_t proto) const
{
  if (!port_list) {
    syslog(LOG_WARNING, "Missing port list");
    return;
  }

  iterator<vector<int>,int> it(*port_list);
  for (it.start(); it.cont(); it.next())
    spawn_server(proto, it.get());
}

/* Eof: proto_map.cc */
